import pygame
import math
import sys
import tkinter as tk
from tkinter import filedialog

from pygame.version import ver

#
#  ####### ####### ##    ## ##   ##  ######   ###### ####### ##      ##      ####### 
#  ##      ##      ##    ## ##  ##  ##    ## ##      ##      ##      ##      ##      
#  ####### #####   ##    ## #####   ##    ## ##      #####   ##      ##      ####### 
#       ## ##      ##    ## ##  ##  ##    ## ##      ##      ##      ##           ## 
#  ####### #######  ######  ##   ##  ######   ###### ####### ####### ####### ####### 
#                                                                                    
#                                                                   by masuo, 2022.07
#                                                                         version 1.1
#                                                                Do not redistribute!

# constants #
_winwidth = 1280
_winheight = 720
_partscale = 10
_pallete = [
    [0,0],
    [255,0],
    [255,255],
    [0,196]
]
_types = [
    "air",
    "wire",
    "gate",
    "input"
]

# classes #
class Cell:
    power = 0
    switched = 0

    def __init__(self, type):
        self.type = type

# functions #

def screenToGridXY(x,y):
    return [math.floor(x/_partscale), math.floor(y/_partscale)]

def prepareGrid(width, height, zsize):
    # prepare a grid of given dimensions
    grid = [[[None for k in range(zsize+1)] for j in range(height+1)] for i in range(width+1)]
    return grid

def drawStripes(screen, start, end, vertical, second):
    for x in range(end, start, 5 if second else -5):
        pygame.draw.rect(
            screen, 
            (128, 128, 0),
            pygame.Rect(
                0 if vertical else x,
                x if vertical else 0,
                _winwidth if vertical else 1,
                1 if vertical else _winheight
            )
        )

def drawRegion(screen, regionFirst, regionSecond):
    drawStripes(screen, 0, regionFirst[0]*_partscale, False, False)
    drawStripes(screen, 0, regionFirst[1]*_partscale, True, False)
    drawStripes(screen, _winwidth, regionSecond[0]*_partscale, False, True)
    drawStripes(screen, _winheight, regionSecond[1]*_partscale, True, True)

def drawGrid(grid, screen, currentLayer, layersVisible):
    layerThickness = math.floor(255/layersVisible)
    # for every cell
    for x in range(len(grid)):
        for y in range(len(grid[x])):
            for z in range(currentLayer+1):
                # if it's not an air
                if grid[x][y][z] != None:
                    # calculate alpha basing on current layer
                    alpha = 255 - layerThickness*abs(currentLayer-z)
                    if alpha < 0: alpha = 0

                    # draw a cell
                    pygame.draw.rect(
                        screen, 
                        pygame.Color.premul_alpha(pygame.Color(
                            *_pallete[grid[x][y][z].type],
                            (int)(grid[x][y][z].power*4), 
                            alpha
                        )), 
                        pygame.Rect(x*_partscale,y*_partscale,_partscale,_partscale)
                    )

def drawText(screen, font, text, color, position):
    selectedTypeText = font.render(text, False, color)
    screen.blit(selectedTypeText, position)

def fillCells(grid, position, z, type, occupiedCells):
    # create cells at given screen position
    x, y = screenToGridXY(*position)
    if type == 0:
        grid[x][y][z] = None
        if [x, y, z] in occupiedCells:
            occupiedCells.remove([x, y, z])
    else:
        grid[x][y][z] = Cell(type)
        if [x, y, z] not in occupiedCells:
            occupiedCells.append([x, y, z])
        
    return grid, occupiedCells

def serializeGrid(grid):
    # write file header
    serialdata = ""
    gridwidth = len(grid)
    gridheight = len(grid[0])
    gridzsize = len(grid[0][0])
    serialdata += "title untitled\nsizes " # title is currently unused
    serialdata += str(gridwidth) + " " + str(gridheight) + " " + str(gridzsize) + "\n"
    serialdata += "cells\n"

    # write "cells" part of the file
    # for every cell
    for x in range(gridwidth):
        for y in range(gridheight):
            for z in range(gridzsize):
                # if it's not an air
                if grid[x][y][z] != None:
                    # save its parameters
                    currentCell = grid[x][y][z]
                    serialdata += str(x) + " " + str(y) + " " + str(z) + " " + str(currentCell.type)
                    serialdata += " " + str(currentCell.power) + " " + str(currentCell.switched)
                    serialdata += " 0 0 0\n"

    return serialdata

def deserializeGrid(serialData):
    # prepare some vars
    grid = []
    occupiedCells = []
    dataLines = serialData.splitlines()
    width = 0
    height = 0
    zsize = 0
    started = False

    # read file by lines
    for line in dataLines:
        # if "cells" part of the file started
        if started == True:
            # assign values and create a cell
            values = [int(x) for x in line.split()]
            newCell = Cell(values[3])
            newCell.power = values[4]
            newCell.switched = values[5]
            grid[values[0]][values[1]][values[2]] = newCell
            occupiedCells.append([values[0], values[1], values[2]]) 

        # if header of the file is currently read
        else:
            # read grid dimensions
            if line.startswith("sizes"):
                width, height, zsize = [int(x)-1 for x in line.split()[1::]]
            # prepare an empty grid of given dimensions if "cells" part of the file is about to start
            if line == "cells" and width > 0:
                started = True
                grid = prepareGrid(width, height, zsize)

    return grid, [width, height, zsize], occupiedCells

def cropGrid(grid, x, y, x2, y2):
    newgrid = grid[x:x2]
    for a in range(len(newgrid)):
        newgrid[a] = newgrid[a][y:y2]
    return newgrid 

def saveGrid(grid):
    # serialize data
    print("Serializing...")
    serialdata = serializeGrid(grid)
    print("Ok!")

    # select path
    f = filedialog.asksaveasfile(mode='w', defaultextension=".seu", filetypes =[('Seuko Files', '*.seu')])
    if f is None: 
        return
    
    # write to a file
    f.write(serialdata)
    f.close()
    print("Saved!")

def loadGrid():
    # select file
    f = filedialog.askopenfile(mode='r', filetypes =[('Seuko Files', '*.seu')])
    if f is None: 
        return None, None, None
    
    # read from file
    print("Reading...")
    serialdata = f.read()
    f.close()

    # deserialize data
    print("Deserializing...")
    grid, sizesInfo, occupiedCells = deserializeGrid(serialdata)
    print("Ok!")
    return grid, sizesInfo, occupiedCells

def switchCell(grid, position, z):
    # if cell isn't an air, switch its mode
    x, y = screenToGridXY(*position)
    if grid[x][y][z] != None:
        grid[x][y][z].switched +=1
    return grid

def doPhysics(grid, occupiedCells):
    # for every occupied cell
    for cell in occupiedCells:
        x, y, z = cell
        if grid[x][y][z] == None:
            occupiedCells.remove([x, y, z])
            continue
    
        neighbourPower = [0]*6
        neighbourPowerI = [0]*6
        neighbourPowerG = [0]*6                

        # for every neighbour
        for i in range(6):
            xco = (x-1+(i*2)) if i < 2 else x
            yco = (y-5+(i*2)) if i > 1 and i < 4 else y
            zco = (z-9+(i*2)) if i > 3 else z
            # if neighbour isn't an air, check its powers
            if grid[xco][yco][zco] != None:
                neighbourPower[i] = grid[xco][yco][zco].power
                if grid[xco][yco][zco].type == 3: neighbourPowerI[i] = neighbourPower[i]
                if grid[xco][yco][zco].type != 2: neighbourPowerG[i] = neighbourPower[i]

        # power transfers
        if grid[x][y][z].type == 1: # WIRE
            grid[x][y][z].power = max(neighbourPower)-1
            if(grid[x][y][z].power < 0): grid[x][y][z].power = 0
            
        if grid[x][y][z].type == 2: # NOT GATE
            grid[x][y][z].power = 63 if max(neighbourPowerI) == 0 else 0
            if grid[x][y][z].switched % 2 == 1:
                grid[x][y][z].power = 63 - grid[x][y][z].power

        if grid[x][y][z].type == 3: # GATE INPUT
            grid[x][y][z].power = max(neighbourPowerG)-1
            if(grid[x][y][z].power < 0): grid[x][y][z].power = 0  

    return grid

# main #
def main():   

    # prepare tkInter #
    root = tk.Tk()
    root.withdraw()
        
    # prepare pygame #
    screen = pygame.display.set_mode((1280,720))
    pygame.display.set_caption('seukoCells')

    Icon = pygame.image.load('favicon.png')
    pygame.display.set_icon(Icon)

    pygame.font.init()
    fontRoboto = pygame.font.SysFont('Roboto', 24)

    clock = pygame.time.Clock()

    # prepare rest
    xsize, ysize = screenToGridXY(_winwidth, _winheight)
    zsize = 3
    grid = prepareGrid(xsize, ysize, zsize)
    occupiedCells = []
    
    selectedType = 0
    currentLayer = 1
    layersVisibile = 1
    pressStartPos = None
    pressDirection = (1,1)
    regionFirst = (0,0)
    regionSecond = (xsize,ysize)
    regionMode = 0
    # 0 - default
    # 1 - region first
    # 2 - region second

    print("Welcome to seukoCells, starting main loop...")

    # main loop
    while True:

        # check for events
        for event in pygame.event.get():
            # close window event
            if event.type == pygame.QUIT:
                sys.exit(0)

            # keyboard events
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    sys.exit(0)
                elif event.key == pygame.K_s:
                    saveGrid(grid)
                elif event.key == pygame.K_l:
                    newgrid, newsizes, newocells = loadGrid()
                    if newgrid != None: 
                        grid = newgrid
                        xsize, ysize, zsize = newsizes
                        occupiedCells = newocells
                        currentLayer = 0
                elif event.key == pygame.K_z:
                    selectedType-=1
                    if selectedType < 0: selectedType = len(_types)-1
                elif event.key == pygame.K_x:
                    selectedType+=1
                    if selectedType >= len(_types): selectedType = 0
                elif event.key == pygame.K_k:
                    layersVisibile+=1
                    if layersVisibile > zsize: layersVisibile = 1
                elif event.key == pygame.K_r:
                    regionMode = 1
                    regionFirst = (0,0)
                    regionSecond = (xsize,ysize)
            
            # mouse events
            elif event.type == pygame.MOUSEBUTTONDOWN:
                if event.button == 1 and regionMode > 0:
                    regionMode += 1
                    if regionMode > 3: 
                        regionMode = 0
                        regionFirst = (0,0)
                        regionSecond = (xsize, ysize)
                if event.button == 3:
                    pos = pygame.mouse.get_pos()
                    grid = switchCell(grid, pos, currentLayer)
                if event.button == 4:
                    if currentLayer < zsize-1: currentLayer+=1
                if event.button == 5:
                    if currentLayer > 0: currentLayer-=1

        # LMB pressed
        if pygame.mouse.get_pressed()[0]:
            pos = pygame.mouse.get_pos()
            if regionMode == 0:
                if pressStartPos != None and pressDirection == (1,1) and abs(pos[1] - pressStartPos[1]) < 2*_partscale:
                    pressVD = abs(pos[0] - pressStartPos[0]) - abs(pos[1] - pressStartPos[1])
                    if pressVD > _partscale:
                        pressDirection = (1,0)
                    elif pressVD < -_partscale:
                        pressDirection = (0,1)
                if pressStartPos == None: 
                    pressStartPos = pos
                posOffset = (pos[0]-pressStartPos[0], pos[1]-pressStartPos[1])
                posLockedDirection = (pressStartPos[0]+posOffset[0]*pressDirection[0], pressStartPos[1]+posOffset[1]*pressDirection[1])
                grid, occupiedCells = fillCells(grid, posLockedDirection, currentLayer, selectedType, occupiedCells)
        else:
            pressStartPos = None
            pressDirection = (1,1)

        if regionMode > 0 and regionMode < 3:
            pos = pygame.mouse.get_pos()
            if regionMode == 1:
                regionFirst = screenToGridXY(*pos)
            elif regionMode == 2:
                regionSecond = screenToGridXY(*pos)

        # physics part
        grid = doPhysics(grid, occupiedCells)

        # screen drawing part
        clock.tick()
        screen.fill((0,0,0))
        drawGrid(grid, screen, currentLayer, layersVisibile)
        drawRegion(screen, regionFirst, regionSecond)
        drawText(screen, fontRoboto, _types[selectedType], (255, 255, 255), (10,10))
        drawText(screen, fontRoboto, "Layer: "+ str(currentLayer) + " (+" + str(layersVisibile-1) + ")", (255, 255, 255), (10,26))
        drawText(screen, fontRoboto, "B: "+ str(len(occupiedCells)) + " FPS: " + str(int(clock.get_fps())), (255, 255, 255), (10,40))
        pygame.display.flip()


main()

# have a wonderful day! #